#include "libgraph.h"
#include <memory.h>
#import <Cocoa/Cocoa.h>

extern NSLock* lock;
extern int width, height;
extern float red, green, blue;
extern BOOL needsRefresh;
extern unsigned char* data;
extern int tailleImage;

extern const double_t mathPi;

value caml_gr_fill_rect(value a, value b, value vw, value vh)
{
    caml_gr_check_open();
    [lock lock];
    int x = Int_val(a);
    int y = Int_val(b);
    int w = Int_val(vw);
    int h = Int_val(vh);
    NSAutoreleasePool* pool =[[NSAutoreleasePool allocWithZone:nil] init];
    NSBezierPath* path = [NSBezierPath bezierPath];
    [path moveToPoint:NSMakePoint(x, y)];
    [path lineToPoint: NSMakePoint(x+w, y)];
    [path lineToPoint:NSMakePoint(x+w, y+h)];
    [path lineToPoint:NSMakePoint(x,y+h)];
    [path lineToPoint:NSMakePoint(x,y)];
    [path fill];
    [pool release];
    needsRefresh = YES;
    [lock unlock];
    return Val_unit;
}

value caml_gr_fill_poly(value array)
{
	int n_points, i;
	n_points = Wosize_val(array);
	NSAutoreleasePool* pool =[[NSAutoreleasePool allocWithZone:nil] init];
	NSBezierPath* path = [NSBezierPath bezierPath];
	[path moveToPoint:NSMakePoint(Int_val(Field(Field(array, 0), 0)), Int_val(Field(Field(array, 0), 1)))];	
	for(i = 1; i < n_points; i++)
		[path lineToPoint:NSMakePoint(Int_val(Field(Field(array, i), 0)), Int_val(Field(Field(array, i), 1)))];
	[path lineToPoint:NSMakePoint(Int_val(Field(Field(array, 0), 0)), Int_val(Field(Field(array, 0), 1)))];
	[lock lock];
	[path fill];
	[pool release];
	needsRefresh = YES;
	[lock unlock];
	return Val_unit;
}

value caml_gr_fill_arc_nat(value vx, value vy, value vrx, value vry, value va1, value va2)
{
	caml_gr_check_open();
	[lock lock];
	NSAutoreleasePool* pool =[[NSAutoreleasePool allocWithZone:nil] init];
	[[NSGraphicsContext currentContext] saveGraphicsState];
	NSBezierPath* path = [NSBezierPath bezierPath];
	float cx = Int_val(vx);
	float cy = Int_val(vy);
	float r1 = Int_val(vrx);
	float r2 = Int_val(vry);
	float a1 = Int_val(va1);
	float a2 = Int_val(va2);
	NSPoint c = NSMakePoint(cx,cy);
	[path moveToPoint: c];
	float r = r1 + r2;
	[path lineToPoint: NSMakePoint(cx + r * cos(mathPi*a1/180.), cy + r * sin(mathPi*a1/180))];
	[path appendBezierPathWithArcWithCenter:c radius:r startAngle:a1 endAngle:a2];
	[path closePath];
	[path addClip];
	[[NSBezierPath bezierPathWithOvalInRect:NSMakeRect(cx - r1, cy -r2, 2 * r1, 2 * r2)] fill];
	[[NSGraphicsContext currentContext] restoreGraphicsState];
	[pool release];
	needsRefresh = YES;
	[lock unlock];
	return Val_unit;
}

value caml_gr_fill_arc(value *argv, int argc)
{
	return caml_gr_fill_arc_nat(argv[0], argv[1], argv[2], argv[3], argv[4], argv[5]);
}
