#include <caml/alloc.h>
#include <caml/signals.h>
#include <unistd.h>
#include "libgraph.h"
#import <Cocoa/Cocoa.h>
#import "MCEvent.h"
#import "MCProtocol.h"

extern NSLock* lock;
extern NSMutableArray* queue;
extern NSDistantObject <ServerProtocol>* mcProxy;
extern int tailleImage;
extern unsigned char* data;

value caml_gr_wait_event(value eventlist)
{
	caml_gr_check_open();

	int mask = 0;
	BOOL poll = NO;
	camlEvent cev;
	cev.mask = 0;
	cev.mouse_x = -1;
	cev.mouse_y = -1;
	cev.key = 0;

    while (eventlist != Val_int(0)) {
        switch (Int_val(Field(eventlist, 0))) {
            case 0:                     // Button_down
                mask |= MOUSEDOWN; break;
            case 1:                     // Button_up
                mask |= MOUSEUP; break;
            case 2:                     // Key_pressed
                mask |= KEYDOWN; break;
            case 3:                     // Mouse_motion
                mask |=MOUSEMOVE; break;
            case 4:                     // Poll
                poll = YES; break;
        }
        eventlist = Field(eventlist, 1);
    }

	if (poll)
	{
        [lock lock];
        int n = 0;
        if (queue != nil && (n =[queue count]))
        {
            int i = 0;
            while (i < n)
            {
                MCEvent* me = [queue objectAtIndex:i];
                cev.mask = [me mask];
                cev.mouse_x = [me mouse_x];
                cev.mouse_y = [me mouse_y];
                cev.key = [me key];
                if ((cev.mask & KEYDOWN) != 0)
                {
                    break;
                }
                i++;
            }
        }
        [lock unlock];
	}
	else
	{
		while ((cev.mask & mask) == 0)
		{
			caml_gr_check_open();
			usleep(20);
            
			caml_enter_blocking_section();
			caml_leave_blocking_section();
            
			[lock lock];
			int n = 0;
			if (queue != nil && (n =[queue count]))
			{
				int i = 0;
				while (i < n)
				{ 
                    MCEvent* me = [queue objectAtIndex:i];
                    cev.mask = [me mask];
                    cev.mouse_x = [me mouse_x];
                    cev.mouse_y = [me mouse_y];
                    cev.key = [me key];
					if ((cev.mask & mask) != 0)
					{
						[queue removeObjectAtIndex:i];
						break;
					}
					i++;
				}
			}
			[lock unlock];
		}
	}
    if (cev.mask & MOUSEMOVE)
    {
		[mcProxy getImage: [NSData dataWithBytes: data length: tailleImage]];
    }
	value res = caml_alloc_tuple(5);
    Field(res, 0) = Val_int(cev.mouse_x);
    Field(res, 1) = Val_int(cev.mouse_y);
	Field(res, 2) = Val_bool(cev.mask & (MOUSEDOWN | MOUSEMOVE));
	Field(res, 3) = Val_bool(cev.mask & KEYDOWN);
	Field(res, 4) = Val_int(cev.key & 0xFF);
	return res;
}
