#import <Cocoa/Cocoa.h>
#import "libgraph.h"
#import "MCEvent.h"
#import "MCProtocol.h"
#include <string.h>
#include <caml/callback.h>
#include <caml/fail.h>
#include <unistd.h>

#undef alloc

/*----------------------------------------------------------------------------------------------------*/

//Variables globales

NSAutoreleasePool *pool = nil;
int tailleImage;
int grx, gry;
int orx, ory;
int width;
int height;
NSBitmapImageRep* bitmap = nil;
unsigned char* data = nil;
float red, green, blue;
NSRect frame;
NSString* title = nil;
NSTimer* timer = nil;
char fontName[255];
float fontSize;
int lineWidth;
BOOL needsRefresh = NO;
BOOL newTitle = NO;
BOOL graphIsOpen = NO;
NSLock* lock = nil;		//synchronisation du thread principal (commandes ralisant le dessin) et du thread
						//secondaire, ce dernier envoyant priodiquement ce qui a t dessin au serveur
						//MacCaml qui gre la fentre graphique

NSDistantObject <ServerProtocol>* mcProxy = nil;	// proxy pour la vue vendue par le serveur MacCaml
NSMutableArray* queue = nil; // reproduit la file des vnements en provenance du serveur MacCaml
char display[20]; // le nom sous lequel est vendue la vue

/*----------------------------------------------------------------------------------------------------*/

//Classe pour un objet grant dans un thread secondaire la relation avec la vue vendue par MacCaml,
//La boucle interactive de Caml fonctionnant dant le thread principal.

@interface MCClient:NSObject
{
	int b;
	NSData* tiffData;
}
- (void) mcThread;
- (void) refresh:(NSTimer*) timer;
@end

@implementation MCClient
- (id) init
{
	if (self = [super init])
	{
		tiffData = nil;
	}
	return self;
}

-(void) refresh:(NSTimer*) tmr
{
	if (b == 1)
	{
		usleep(100000);
		b = 0;
	}
	[lock lock];
	if (graphIsOpen == NO)
	{
		if (data) {free(data); data = nil;}
		[bitmap release];
		bitmap = nil;
		[timer invalidate];
		timer = nil;
		[lock unlock];
		[queue release];
		queue = nil;
		return;
	}
	if (newTitle == YES)
	{		
		[mcProxy setServerFrame:frame title:title];
		newTitle = NO;
		needsRefresh = YES;
	}
	if (needsRefresh == YES)
	{
		[mcProxy getImage: [NSData dataWithBytes: data length: tailleImage]];
		needsRefresh = NO;
	}

	MCEvent* e = nil;
l:	e = [mcProxy getNextEvent];
	if (e == nil)
	{
		[lock unlock];
		return;
	}
    if ([e mask] == WINDOWCLOSED)
	{
		graphIsOpen = NO;
		needsRefresh = NO;
		[mcProxy getImage: nil];
		[lock unlock]; 
		return;
	}
	if (queue != nil)
	{
        MCEvent* ee = [[MCEvent allocWithZone:nil]init];
        [ee setMask:[e mask]];
        [ee setMouse_x:[e mouse_x]];
        [ee setMouse_y:[e mouse_y]];
        [ee setKey:[e key]];
		[queue addObject:ee];
        [lock unlock];
        usleep(20000);
        [lock lock];
		goto l;
	}
	
	[lock unlock];
}

- (void) mcThread
{
	[lock lock];
	NSAutoreleasePool *p = [[NSAutoreleasePool allocWithZone:nil] init];
	
	NSString* serverName = [NSString stringWithCString:display encoding: NSMacOSRomanStringEncoding];
	mcProxy = (NSDistantObject<ServerProtocol>*) [NSConnection
		rootProxyForConnectionWithRegisteredName:serverName
											host:nil];
	if (mcProxy == nil)
	{
		graphIsOpen = NO;
		[p release];
		[lock unlock];
		return;
	}
	
	[[mcProxy connectionForProxy] enableMultipleThreads];
	
	[(NSDistantObject*)mcProxy setProtocolForProxy:@protocol(ServerProtocol)];
	[mcProxy setServerFrame:frame title:title];
	
	b = 1;
	timer = [NSTimer scheduledTimerWithTimeInterval:0.1 target:self selector:@selector(refresh:)
										   userInfo:nil repeats:YES];
	
	[[NSRunLoop currentRunLoop] addTimer:timer forMode:NSDefaultRunLoopMode];
	[lock unlock];
	
	[[NSRunLoop currentRunLoop]run];
	
	[lock lock];
	[p release];
	[lock unlock];
}
@end

/*----------------------------------------------------------------------------------------------------*/

// Fonctions graphiques de Caml

MCClient* mcClient = nil; //l'objet charg de la relation avec MacCaml

void setTitle(NSString* s)
{
  [s retain];
  [title release];
  title = s;
}

static value * graphic_failure_exn = NULL;

void caml_gr_fail(char *fmt, char *arg)
{
  char buffer[1024];
  if (graphic_failure_exn == NULL) {
    graphic_failure_exn = caml_named_value("Graphics.Graphic_failure");
    if (graphic_failure_exn == NULL)
      caml_invalid_argument("Exception Graphics.Graphic_failure not initialized, must link graphics.cma");
  }
  sprintf(buffer, fmt, arg);
  caml_raise_with_string(*graphic_failure_exn, buffer);
}

void caml_gr_check_open(void)
{
	[lock lock];
	BOOL b = graphIsOpen;
	[lock unlock];
	if (b == NO)
	{
		[lock lock];
		if (data) {free(data); data = nil;}
		[bitmap release];
		bitmap = nil;
		[timer invalidate];
		timer = nil;
		[lock unlock];
		[queue release];
		queue = nil;
		[pool release];
		pool = nil;
		caml_gr_fail("graphic screen not opened", "");
	}
} 

value caml_gr_open_graph(value s)
{	
	if (lock == nil) lock = [[NSLock allocWithZone: nil] init];

	[lock lock];
	if (graphIsOpen == YES) { [lock unlock]; return Val_unit; }
	graphIsOpen = YES;
	[lock unlock];
	
	orx = 0;
	ory = 0;
	width = 600;
	height = 450;

	sscanf(String_val(s), "%dx%d+%d+%d", &width, &height, &orx, &ory);
	frame = NSMakeRect(orx, ory, width, height);


	if (pool == nil) pool = [[NSAutoreleasePool allocWithZone: nil] init];
	NSApp = [NSApplication sharedApplication];

	setTitle(@"Caml Graphics");

	strcpy(display, getenv("CAMLSERVERNAME"));

	queue =[[NSMutableArray allocWithZone:nil] initWithCapacity:100];
	tailleImage = width*height*4;
	data = (unsigned char*)malloc(tailleImage);
	bitmap = [[NSBitmapImageRep allocWithZone:nil] 
							initWithBitmapDataPlanes:&data
										  pixelsWide:width
										  pixelsHigh:height
									   bitsPerSample:8   
									 samplesPerPixel:4 hasAlpha:YES
											isPlanar:NO
									  colorSpaceName:NSCalibratedRGBColorSpace
										 bytesPerRow:(4 * width)
										bitsPerPixel:32];
	NSGraphicsContext* gc = nil;
	gc = [NSGraphicsContext graphicsContextWithBitmapImageRep:bitmap];
	[NSGraphicsContext setCurrentContext:gc];
	[gc setShouldAntialias:NO];
	[[NSColor whiteColor] set];
	NSRectFill(NSMakeRect(0,0,width,height));
	[[NSColor blackColor] set];
	[[NSFont fontWithName:@"Courier" size:12] set];
	strcpy(fontName, "Courier");
	needsRefresh = YES;
	newTitle = NO;
	timer = nil;
	grx = 0;
	gry = 0;
	fontSize = 12;
	lineWidth = 0;

	if (mcClient == nil) mcClient = [[MCClient allocWithZone:nil] init];
	
	[NSThread detachNewThreadSelector:@selector(mcThread) toTarget:mcClient withObject:nil];
	return Val_unit;
}

value caml_gr_close_graph(void)
{
	[lock lock];
	if (graphIsOpen == NO){[lock unlock]; return Val_unit;}
	graphIsOpen = NO;
	needsRefresh = NO;
	
	if (data) 
	{
		[[NSColor whiteColor] set];
		NSRectFill(NSMakeRect(0,0,frame.size.width,frame.size.height));
		[mcProxy getImage: [NSData dataWithBytes: data length: tailleImage]];
		if (data) {free(data); data = nil;}		
	}
	
	[bitmap release];
	bitmap = nil;
	[mcProxy hideWindow];
	[timer invalidate];
	timer = nil;
	free(data);
	[queue release];
	queue = nil;
    [title release];
    title = nil;
	[pool release];
	pool = nil;
	
	[lock unlock];
	return Val_unit;
}

value caml_gr_window_id(void)
{
  return Val_unit;
}

value caml_gr_open_subwindow(value vx, value vy, value width, value height)
{
    return Val_unit;
}

value caml_gr_close_subwindow(value wid)
{
    return Val_unit;
}

value caml_gr_set_window_title(value s)
{
	[lock lock];
	setTitle([NSString stringWithCString:String_val(s) encoding:NSUTF8StringEncoding]);
	newTitle = YES;
	[lock unlock];
	return Val_unit;
}

value caml_gr_resize_window (value vx, value vy)
{
  caml_gr_check_open();
	[lock lock];
	if (graphIsOpen == NO){[lock unlock]; return Val_unit;}
	
	if (data) 
	{
		[[NSColor whiteColor] set];
		NSRectFill(NSMakeRect(0,0,frame.size.width,frame.size.height));
		[mcProxy getImage: [NSData dataWithBytes: data length: tailleImage]];
		if (data) {free(data); data = nil;}		
	}
	
	[bitmap release];
	bitmap = nil;
	[mcProxy hideWindow];
	free(data);
	[queue release];
	queue = nil;
	[pool release];
	pool = nil;
	
	width = Int_val(vx);
	height = Int_val(vy);
	frame = NSMakeRect(orx, ory, width, height);

	title = @"Caml Graphics";
	if (pool == nil) pool = [[NSAutoreleasePool allocWithZone: nil] init];
	queue =[[NSMutableArray allocWithZone:nil] initWithCapacity:100];
	tailleImage = width*height*4;
	data = (unsigned char*)malloc(tailleImage);
	bitmap = [[NSBitmapImageRep allocWithZone:nil] 
							initWithBitmapDataPlanes:&data
										  pixelsWide:width
										  pixelsHigh:height
									   bitsPerSample:8   
									 samplesPerPixel:4 hasAlpha:YES
											isPlanar:NO
									  colorSpaceName:NSCalibratedRGBColorSpace
										 bytesPerRow:(4 * width)
										bitsPerPixel:32];
	NSGraphicsContext* gc = nil;
	gc = [NSGraphicsContext graphicsContextWithBitmapImageRep:bitmap];
	[NSGraphicsContext setCurrentContext:gc];
	[gc setShouldAntialias:NO];
	[[NSColor whiteColor] set];
	NSRectFill(NSMakeRect(0,0,width,height));
	[[NSColor blackColor] set];
	[[NSFont fontWithName:@"Courier" size:10] set];
	strcpy(fontName, "Courier");
	needsRefresh = YES;
	newTitle = NO;
	timer = nil;
	grx = 0;
	gry = 0;
	fontSize = 10;
	lineWidth = 0;	
	[lock unlock];
	[mcProxy setServerFrame:frame title:title];
  return Val_unit;
}

value caml_gr_clear_graph(void)
{
	caml_gr_check_open();
	[lock lock];
	[[NSColor whiteColor] set];
	NSRectFill(NSMakeRect(0,0,frame.size.width,frame.size.height));
	[[NSColor colorWithCalibratedRed: red green: green blue: blue alpha: 1] set]; 
	needsRefresh = YES;
	[lock unlock];
	return Val_unit;
}

value caml_gr_size_x(void)
{
	caml_gr_check_open();
	int w = frame.size.width;
	return Val_int(w);
}

value caml_gr_size_y(void)
{
	caml_gr_check_open();
	int h = frame.size.height;
	return Val_int(h);
}

value caml_gr_synchronize(void)
{
  caml_gr_check_open();
  return Val_unit ;
}

value caml_gr_display_mode(value flag)
{
  caml_gr_check_open();
  return Val_unit ;
}

value caml_gr_remember_mode(value flag)
{
  caml_gr_check_open();
  return Val_unit ;
}

value caml_gr_sigio_signal(value unit)
{
	return Val_int(EVENT_SIGNAL);
}

value caml_gr_sigio_handler(void)
{
  return Val_unit;
}
